import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from .migr_simulator import MigrSimulator

from .hmm_simulator import PotentialType


class WifiSimulator(MigrSimulator):
    """simulate ensemble move, sensor is sparse
    - place sensors in some place
    - draw various data
    """

    def __init__(self, time_step, d_col, d_row, random_seed):
        # use random_seed 3, random_sensor 16, reproduce the experiment result
        super().__init__(time_step, d_col, d_row, random_seed)
        self.record.obser_num = 0
        self.hotspot = []

    def compile(self):
        super().compile()

        potential = np.zeros(self.record.state_num)
        potential[0] = 0.7
        potential[int(self.d_col / 2)] = 0.3
        self.register_potential(PotentialType.INIT, potential)

    def register_hotspot(self, row, col):
        self.record.obser_num += 1
        self.hotspot.append((row, col))

    def _produce_sensor_potential(self):
        potential = []
        for cur_row in range(self.d_row):
            for cur_col in range(self.d_col):
                cur_potential = np.zeros(self.record.obser_num)

                for i, sensor in enumerate(self.hotspot):
                    distance = np.linalg.norm(
                        [cur_row - sensor[0], cur_col - sensor[1]])
                    cur_potential[i] = np.min([0.99, 1.0 * np.exp(-distance)])

                potential.append(cur_potential / np.sum(cur_potential))

        self.register_potential(
            PotentialType.EMISSION,
            np.array(potential).reshape(
                self.record.state_num,
                self.record.obser_num))

        return int(self.d_col / 2)

    def fixed_sensors(self, space_d=1):
        for cur_row in range(0, self.d_row, space_d):
            for cur_col in range(0, self.d_col, space_d):
                self.register_hotspot(cur_row, cur_col)

    def random_sensor(self, num):
        for _ in range(num):
            row = self.rng.uniform(0, self.d_row - 1)
            col = self.rng.uniform(0, self.d_col - 1)
            self.register_hotspot(row, col)

    def draw_sensor(self):
        """draw wifi sensor position in grid
        """
        sensor_row = []
        sensor_col = []
        for sensor in self.hotspot:
            sensor_row.append(sensor[0] + 0.2)
            sensor_col.append(sensor[1] + 0.2)

        marker_col, marker_row = np.meshgrid(
            np.arange(self.d_col),
            np.arange(self.d_row))

        pos_x = np.concatenate((marker_col.flatten(), sensor_col), axis=0)
        pos_y = np.concatenate((marker_row.flatten(), sensor_row), axis=0)
        x_y_size = np.concatenate(
            (np.ones(int(marker_col.size)) * 1.0, np.ones(len(sensor_col)) * 5.0))
        style = ['cell'] * marker_col.size + ['sensor'] * len(sensor_row)
        ax = sns.scatterplot(x=pos_y,
                             y=pos_x,
                             size=x_y_size,
                             hue=style,
                             style=style,
                             markers={'cell': 's', 'sensor': 'X'})

        ax.set_xlim([-1, self.d_col])
        ax.set_xticks(np.arange(0, self.d_col, 2))
        ax.set_xticklabels(np.arange(0, self.d_col, 2).tolist())
        ax.set_ylim([-1, self.d_row])
        ax.set_yticks(np.arange(0, self.d_row, 2))
        ax.set_yticklabels(np.arange(0, self.d_row, 2).tolist())
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, ["cell", "sensor"])
        plt.savefig(f"{self.path}/sensor.png", bbox_inches='tight')
        plt.close()

    def viz_sensor(self):
        """draw raw sensor data
        """
        for i in range(self.record.time_step):
            locations = self.record["sensor"][:, i]
            bins, _ = np.histogram(
                locations, np.arange(
                    self.record.obser_num + 1))
            self.__draw_sensordata(bins, f"{self.path}/observer_{i}_step.png")

    def __draw_sensordata(self, bins, fig_name):
        xx = []
        yy = []
        xy_size = []
        for xy, cnt in enumerate(bins):
            if cnt > 0:
                xx.append(self.hotspot[xy][0])
                yy.append(self.hotspot[xy][1])
                xy_size.append(int(cnt))
        self.visualizer.visualize_location(xx, yy, xy_size, **{
            "fig_name": fig_name,
            "xlabel": 'Observed'})



from dataclasses import dataclass


@dataclass
class BaseConfig:

    @staticmethod
    def itsbp_schedule(cnt, leaf_nodes):
        cnt += 1
        return cnt % len(leaf_nodes)


baseconfig = BaseConfig()



import unittest
import numpy as np
from cbp.builder import HMMSimulator, PotentialType


class TestSim(unittest.TestCase):
    def test_sim_sample(self):
        rng = np.random.RandomState(1)
        for i in range(10):
            state_num = 10
            obser_num = 5
            sim = HMMSimulator(10, state_num, obser_num, i, is_theorem=False)
            sim.register_potential(
                PotentialType.INIT, rng.dirichlet(
                    [1] * state_num))
            sim.register_potential(
                PotentialType.TRANSITION, rng.dirichlet(
                    [1] * state_num, size=state_num))
            sim.register_potential(
                PotentialType.EMISSION, rng.dirichlet(
                    [1] * obser_num, size=state_num))
            sim.sample(5000)
            sim.record.cal_theory(False)
            self.assertGreater(0.5, sim.record["sim_state_error"])
            self.assertGreater(0.5, sim.record["sim_obser_error"])



import unittest

import numpy as np
from cbp.builder import HMMBuilder, HMMZeroBuilder, LineBuilder
from cbp.graph.coef_policy import bp_policy, avg_policy, factor_policy

from .utils import sinkhorn_bp_equal


class TestGraph(unittest.TestCase):
    def test_hmm_policy(self):
        rng = np.random.RandomState(1)
        for i in range(5):
            for policy in [bp_policy, avg_policy, factor_policy]:
                num_node = int(rng.randint(2, 6))
                node_dim = int(rng.randint(4, 6))
                print(
                    f"Run {i}-th experiment, {num_node} nodes with {node_dim} status")
                self.graph = HMMBuilder(num_node, node_dim, policy)()
                self.graph.run_cnp()
                self.graph.sinkhorn()
                self.assertTrue(all(sinkhorn_bp_equal(self.graph, num_node)))

    def test_line_policy(self):
        rng = np.random.RandomState(1)
        for i in range(5):
            for policy in [bp_policy, avg_policy, factor_policy]:
                num_node = int(rng.randint(2, 4))
                node_dim = int(rng.randint(2, 5))
                print(
                    f"Run {i}-th experiment, {num_node} nodes with {node_dim} status")
                self.graph = LineBuilder(num_node, node_dim, avg_policy)()
                self.graph.run_cnp()
                self.graph.sinkhorn()
                self.assertTrue(all(sinkhorn_bp_equal(self.graph, num_node)))



from .base_builder import BaseBuilder


class TreeBuilder(BaseBuilder):
    def __init__(self, tree_depth, d, policy, rand_seed=1):
        self.tree_depth = tree_depth
        super().__init__(d, policy, rand_seed)

    def init_graph(self):
        self.add_trivial_node()
        for depth in range(self.tree_depth - 2):
            for cur_node in range(2**depth - 1, 2**(depth + 1) - 1):
                for _ in range(2):
                    self.add_branch(head_node=f"VarNode_{cur_node:03d}")

        depth = self.tree_depth - 2
        for cur_node in range(2**depth - 1, 2**(depth + 1) - 1):
            for _ in range(2):
                self.add_branch(head_node=f"VarNode_{cur_node:03d}",
                                is_constrained=True)



from abc import ABC, abstractmethod

import numpy as np
from numpy.random import RandomState
from cbp.graph import GraphModel
from cbp.node import VarNode, FactorNode

from cbp.builder.potential_utils import diagonal_potential_different
from .potential_utils import diagonal_potential, diagonal_potential_conv


class BaseBuilder(ABC):
    def __init__(self, dim, policy, rand_seed=1):
        self.graph = GraphModel(True, coef_policy=policy)
        self.node_dim = dim
        self.rng = RandomState(rand_seed)

    def __call__(self):
        self.init_graph()
        return self.graph

    def add_constrained_node(self, probability=None):
        if probability is None:
            log_probability = self.rng.normal(size=self.node_dim)
            probability = np.exp(log_probability)
        else:
            probability = np.array(probability)

        dim = probability.shape[0]
        varnode = VarNode(dim,
                          constrained_marginal=probability /
                          np.sum(probability))
        self.graph.add_varnode(varnode)
        return varnode

    def add_trivial_node(self, dim=None):
        if dim is None:
            dim = self.node_dim
        varnode = VarNode(dim)
        self.graph.add_varnode(varnode)
        return varnode

    def add_factor(self, name_list, is_conv=False):
        if is_conv:
            factor_potential = diagonal_potential_conv(
                self.node_dim, self.node_dim, self.rng)
        else:
            factor_potential = diagonal_potential(
                self.node_dim, self.node_dim, self.rng)
        factornode = FactorNode(name_list, factor_potential)
        self.graph.add_factornode(factornode)
        return factornode

    def add_factor_different(self, name_list, is_conv=False):
        if is_conv:
            factor_potential = diagonal_potential_conv(
                self.node_dim, self.node_dim, self.rng)
        else:
            factor_potential = diagonal_potential_different(
                self.node_dim, self.node_dim, self.rng)
        factornode = FactorNode(name_list, factor_potential)
        self.graph.add_factornode(factornode)
        return factornode

    def add_branch(self, head_node=None, is_constrained=False,
                   prob=None, is_conv=False):
        if head_node is None:
            head_node = f"VarNode_{self.graph.cnt_varnode-1:03d}"
        if is_constrained:
            node = self.add_constrained_node(prob)
        else:
            node = self.add_trivial_node()

        name_list = [head_node, node.name]
        self.add_factor(name_list, is_conv)

    @abstractmethod
    def init_graph(self):
        pass



import unittest

import numpy as np

from .utils import three_node_tree


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.graph = three_node_tree()

    def test_marginal_bp(self):
        self.graph.run_bp()
        self.graph.exact_marginal()

        isequal_list = []

        for item in self.graph.varnode_recorder.values():
            marginal_equal = np.isclose(item.bfmarginal, item.marginal())
            isequal_list.append(all(marginal_equal))

        self.assertTrue(all(isequal_list))

    def test_marginal_brutal_force(self):
        self.graph.exact_marginal()
        node_equal = np.isclose(np.array([70.0 / 215, 145.0 / 215]),
                                self.graph.get_node("VarNode_000").bfmarginal)
        self.assertTrue(all(node_equal))
        node_equal = np.isclose(np.array([120.0 / 215, 95.0 / 215]),
                                self.graph.get_node("VarNode_001").bfmarginal)
        self.assertTrue(all(node_equal))

        node_equal = np.isclose(np.array([55.0 / 215, 160.0 / 215]),
                                self.graph.get_node("VarNode_002").bfmarginal)
        self.assertTrue(all(node_equal))



import warnings
from functools import partial

import numpy as np
from cbp.node import VarNode
from cbp.utils import (Message, diff_max_marginals,
                       engine_loop)
from cbp.utils.np_utils import (nd_expand, nd_multiexpand,
                                reduction_ndarray)
from cbp.configs.base_config import baseconfig
from .coef_policy import bp_policy
from .graph_utils import cal_marginal_from_tensor
try:
    import pygraphviz  # noqa
except BaseException:
    pygraphviz = None


class BaseGraph():  # pylint: disable=too-many-instance-attributes
    def __init__(self, silent=True, epsilon=1,
                 coef_policy=bp_policy, config=baseconfig):
        self.varnode_recorder = {}
        self.constrained_names = []
        self.leaf_nodes = []
        self.factornode_recorder = {}
        self.node_recorder = {}
        self.epsilon = epsilon
        self.coef_policy = coef_policy
        self.cnt_varnode = 0
        self.cnt_factornode = 0
        self.cfg = config

        # debug utils
        self.silent = silent

    def add_varnode(self, node):
        """add one `~cbp.node.VarNode` to this graph, idx follow the increasing
        order

        :param node: one VarNode
        :type node: VarNode
        :return: name of varnode
        :rtype: str
        """
        assert isinstance(node, VarNode)
        varnode_name = f"VarNode_{self.cnt_varnode:03d}"
        node.format_name(varnode_name)
        self.varnode_recorder[varnode_name] = node
        self.node_recorder[varnode_name] = node
        if node.isconstrained:
            self.constrained_names.append(varnode_name)

        self.cnt_varnode += 1
        return varnode_name

    def add_factornode(self, factornode):
        """add one factor node to the graph
        Do the following tasks

        * add node to the recorders
        * set connections
        * set parent relation

        :param factornode: one factor node
        :type factornode: FactorNode
        :return: name of factor node
        :rtype: str
        """
        factornode.check_potential(self.varnode_recorder)
        factornode_name = f"FactorNode_{self.cnt_factornode:03d}"
        factornode.format_name(factornode_name)
        self.factornode_recorder[factornode_name] = factornode
        self.node_recorder[factornode_name] = factornode

        self.__register_connection(factornode)
        self.__set_parent(factornode)

        self.cnt_factornode += 1
        return factornode_name

    def __register_connection(self, factornode):
        for varnode_name in factornode.get_connections():
            varnode = self.varnode_recorder[varnode_name]
            varnode.register_connection(factornode.name)

    def __set_parent(self, factornode):
        connections = factornode.get_connections()
        factornode.parent = self.varnode_recorder[connections[0]]
        for varnode_name in connections[1:]:
            varnode = self.varnode_recorder[varnode_name]
            varnode.parent = factornode

    def pmf(self):
        """output the probability mass matrix through brutal-force methods

        :return: joint probability mass matrix
        :rtype: ndarray
        """
        varnode_names = list(self.varnode_recorder.keys())
        varnodes = list(self.varnode_recorder.values())
        var_dim = [variable.rv_dim for variable in varnodes]
        assert len(var_dim) < 32, "max number of vars for brute_force is 32 \
            (numpy matrix dim limit)"
        joint_acc = np.ones(var_dim)
        for factor in self.factornode_recorder.values():
            which_dims = [varnode_names.index(v)
                          for v in factor.get_connections()]
            factor_acc = np.ones(var_dim)

            factor_acc = nd_multiexpand(factor.potential, var_dim, which_dims)

            joint_acc *= factor_acc

        joint_prob = joint_acc / np.sum(joint_acc)
        return joint_prob

    def exact_marginal(self):
        varnodes = list(self.varnode_recorder.values())
        prob_tensor = self.pmf()

        marginal_list = cal_marginal_from_tensor(prob_tensor, varnodes)
        for node, marginal in zip(varnodes, marginal_list):
            node.bfmarginal = marginal

    def first_belief_propagation(self):
        for node in self.nodes:
            for recipient_name in node.connections:
                recipient = self.node_recorder[recipient_name]
                if recipient.name not in node.message_inbox:
                    val = node.make_init_message(recipient_name)
                    message = Message(node, val)
                    self.node_recorder[recipient_name].store_message(message)

    def copy_bp_initialization(self, another_graph):
        """copy message setup from the another graph has same structure

        :param another_graph: another graph which close to the optimal point
        :type another_graph: BaseGraph
        """
        # TODO: copy safety!!!
        for node in self.nodes:
            if node.name in another_graph.node_recorder:
                another_node = another_graph.node_recorder[node.name]
                node.message_inbox = another_node.message_inbox
                node.latest_message = another_node.latest_message
            else:
                raise f"{node.name} not in this graph"

    def __init_sinkhorn_node(self):
        varnode_names = list(self.varnode_recorder.keys())
        self.sinkhorn_node_coef = {}  # pylint: disable=attribute-defined-outside-init
        for node_name in self.constrained_names:
            node_instance = self.varnode_recorder[node_name]
            self.sinkhorn_node_coef[node_name] = {
                'index': varnode_names.index(node_name),
                'mu': node_instance.constrained_marginal,
                'u': np.ones(node_instance.rv_dim)
            }
        for node in self.varnode_recorder.values():
            node.sinkhorn = np.ones(node.rv_dim) / node.rv_dim

    def __build_big_u(self):
        varnodes = list(self.varnode_recorder.values())
        var_dim = [variable.rv_dim for variable in varnodes]
        joint_acc = np.ones(var_dim)

        for _, recoder in self.sinkhorn_node_coef.items():
            constrained_acc = nd_expand(
                recoder['u'], tuple(var_dim), recoder['index'])
            joint_acc *= constrained_acc

        # log_joint_acc -= np.max(log_joint_acc)
        return joint_acc / np.sum(joint_acc)

    # TODO this is a bug!!!!
    def sinkhorn_update(self, tilde_c):
        for _, recorder in self.sinkhorn_node_coef.items():
            big_u = self.__build_big_u()
            normalized_denominator = (big_u * tilde_c) / \
                np.sum(big_u * tilde_c)

            copy_denominator = reduction_ndarray(
                normalized_denominator, recorder['index'])
            copy_denominator = np.clip(copy_denominator, 1e-12, None)
            recorder['u'] = recorder['u'] * recorder['mu'] / copy_denominator

        varnodes = list(self.varnode_recorder.values())
        marginal_list = cal_marginal_from_tensor(
            normalized_denominator, varnodes)
        for node, marginal in zip(varnodes, marginal_list):
            node.sinkhorn = marginal

    def __check_sinkhorn(self):
        if len(self.constrained_names) == 0:
            raise RuntimeError(
                "There is no constrained nodes, use brutal force")

    def sinkhorn(self, max_iter=5000000, tolerance=1e-5):
        self.__check_sinkhorn()
        tilde_c = self.pmf()
        self.__init_sinkhorn_node()

        sinkhorn_func = partial(self.sinkhorn_update, tilde_c)
        return engine_loop(engine_fun=sinkhorn_func,
                           max_iter=max_iter,
                           tolerance=tolerance,
                           error_fun=diff_max_marginals,
                           meassure_fun=self.export_sinkhorn,
                           isoutput=False,
                           silent=self.silent)

    def bake(self):
        self.init_node_recorder()
        for node in self.nodes:
            if len(node.connections) == 1:
                root = node
        self.__cal_node_coef(root)

    def __cal_node_coef(self, node):
        node.is_traversed = True
        for item in node.connections:
            if not self.node_recorder[item].is_traversed:
                self.__cal_node_coef(self.node_recorder[item])

        node.auto_coef(self.node_recorder, self.coef_policy)
        node.is_traversed = False

    def init_node_recorder(self):
        factors = list(self.factornode_recorder.values())
        variables = list(self.varnode_recorder.values())
        # in Norm-Product, run factor message first
        self.nodes = factors + variables  # pylint: disable=attribute-defined-outside-init
        self.leaf_nodes = [
            node for node in self.nodes if len(node.get_connections()) == 1]

    def get_node(self, name_str):
        if name_str not in self.node_recorder:
            raise RuntimeError(f"{name_str} is illegal, not in this graph")
        return self.node_recorder[name_str]

    def cal_bethe(self, margin):
        """calculate bethe energy

        :param margin: node_name : margin
        :type margin: dict
        :return: KL divergence between expoert joint dist and p_graph
        :rtype: float
        """
        sum_item = []
        for node in self.nodes:
            sum_item.append(node.cal_bethe(margin[node.name]))

        return np.sum(sum_item)

    def delete_node(self, name_str):
        """delete node from graph, needs to check following

        * judge type, Factor or Variable
        * delete from the various recorders
        * clear connections

        :param name_str: [description]
        :type name_str: [type]
        :raises RuntimeError: [description]
        """
        if name_str not in self.node_recorder:
            raise RuntimeError(f"{name_str} is illegal, not in this graph")
        target_node = self.node_recorder[name_str]
        if isinstance(target_node, VarNode):
            warnings.warn(f"Delete {name_str}, may have a suspend factor node")
        for connected_name in target_node.get_connections():
            connected_node = self.node_recorder[connected_name]
            if connected_node.parent is target_node:
                connected_node.parent = None
            connected_node.get_connections().remove(name_str)

            # clear map
            if len(connected_node.get_connections()) == 0:
                self.__delete_node_recorder(connected_node)

        self.__delete_node_recorder(target_node)

    def __delete_node_recorder(self, node):
        target_map = self.varnode_recorder if isinstance(
            node, VarNode) else self.factornode_recorder
        del target_map[node.name]
        del self.node_recorder[node.name]
        if node.name in self.constrained_names:
            self.constrained_names.remove(node.name)

    def set_node(self, node_name, potential=None, isconstrained=None):
        """change node property
        1. check whether or not in recorder
        2. change potential easily[this may a duplicate function]
        3. change isconstrained if possible, delete from recorder

        :param node_name: [description]
        :type node_name: [type]
        :param potential: [description], defaults to None
        :type potential: [type], optional
        :param isconstrained: [description], defaults to None
        :type isconstrained: [type], optional
        :raises RuntimeError: [description]
        """
        if node_name not in self.node_recorder:
            raise RuntimeError
        node = self.node_recorder[node_name]
        if potential is not None:
            # TODO: make potential property check, when do set
            node.potential = potential
        if isconstrained is not None:
            if node.isconstrained != isconstrained:
                node.isconstrained = isconstrained
                if isconstrained:
                    self.constrained_names.append(node_name)
                else:
                    self.constrained_names.remove(node_name)

    def export_marginals(self):
        """export the marginal for variable nodes

        :return: {node.key:node.marginal}
        :rtype: dict
        """
        return {
            n.name: n.marginal() for n in self.varnode_recorder.values()
        }

    def export_convergence_marginals(self):
        """export the marginal for variable nodes and factor nodes

        :return: {node.key:node.marginal}
        :rtype: dict
        """
        return {n.name: n.marginal() for n in self.nodes}

    def export_sinkhorn(self):
        return {node_name: node.sinkhorn
                for node_name, node in self.varnode_recorder.items()}

    def plot(self, png_name='file.png'):
        """plot the graph through graphviz

            * red: constrained variable node
            * blue: free variable node
            * green: factor

        :param png_name: name of figure, defaults to 'file.png'
        :type png_name: str, optional
        :raises ValueError: [description]
        """
        if pygraphviz is not None:
            graph = pygraphviz.AGraph(directed=False)
            for varnode_name in self.varnode_recorder:
                if varnode_name in self.constrained_names:
                    graph.add_node(varnode_name, color='red', style='filled')
                else:
                    graph.add_node(varnode_name, color='blue', style='bold')

            for name, factornode in self.factornode_recorder.items():
                graph.add_node(name, color='green')
                for varnode_name in factornode.get_connections():
                    graph.add_edge(name, varnode_name)

            graph.layout(prog='neato')
            graph.draw(png_name)
        else:
            raise ValueError("must have pygraphviz installed for visualization")

    def tree_bp(self):
        """run classical belief propagation on a tree graph, only need forward
        and backward

            * add attr: is_send_forward: begin send forward false, after forward
             before backward true, after backward false
        :raises RuntimeError: Only works for the tree graph, loopy graph does
        not work, root node not decided
        """
        assert len(self.constrained_names) == 0
        self.bake()
        self.first_belief_propagation()
        for node in self.nodes:
            node.is_send_forward = False

        tree_root = None
        for node in self.nodes:
            if len(node.connections) == 1:
                tree_root = node
        if tree_root is None:
            raise RuntimeError("graph contains circle")

        self._send_forward(tree_root)
        self._send_backward(tree_root)

    def _send_forward(self, node):
        node.is_send_forward = True
        for cur_node in node.connected_nodes.values():
            if not cur_node.is_send_forward:
                self._send_forward(cur_node)
                cur_node.send_message(node)

    def _send_backward(self, node):
        node.is_send_forward = False
        for cur_node in node.connected_nodes.values():
            if cur_node.is_send_forward:
                node.send_message(cur_node)
                self._send_backward(cur_node)



import unittest

import numpy as np
from cbp.graph import GraphModel
from cbp.graph.coef_policy import bp_policy
from cbp.node import FactorNode, VarNode


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.graph = GraphModel(coef_policy=bp_policy)

        # init varnode
        potential = np.ones(2)
        self.varnode_1 = VarNode(2, potential)
        self.graph.add_varnode(self.varnode_1)

        varnode_name = [node.name for node in [self.varnode_1]]
        factor_potential = np.array(
            [3, 2]
        )
        self.factornode_1 = FactorNode(varnode_name, factor_potential)
        self.graph.add_factornode(self.factornode_1)

        factor_potential_2 = np.array(
            [1, 4]
        )
        self.factornode_2 = FactorNode(varnode_name, factor_potential_2)
        self.graph.add_factornode(self.factornode_2)

    def test_marginal_brutal_force(self):
        self.graph.exact_marginal()
        node_equal = np.isclose(np.array([3.0 / 11, 8.0 / 11]),
                                self.graph.get_node("VarNode_000").bfmarginal)
        self.assertTrue(all(node_equal))

    def test_marginal_bp(self):
        self.graph.run_bp()
        node_equal = np.isclose(np.array([3.0 / 11, 8.0 / 11]),
                                self.graph.get_node("VarNode_000").marginal())
        self.assertTrue(all(node_equal))



from cbp.utils.np_utils import reduction_ndarray


def get_node2root(node):
    rtn = []
    tmp = node
    while True:
        rtn.append(tmp)
        tmp = tmp.parent
        if not tmp:
            break

    return rtn


def find_link(node_a, node_b):
    a_2root = get_node2root(node_a)
    b_2root = get_node2root(node_b)
    while (len(b_2root) > 1 and len(a_2root) > 1):
        if b_2root[-1] == a_2root[-1] and b_2root[-2] == a_2root[-2]:
            b_2root.pop()
            a_2root.pop()
        else:
            break
    b_2root.reverse()
    if len(b_2root) >= 2:
        return a_2root + b_2root[1:]

    return a_2root[:-1] + b_2root[:]


def itsbp_inner_loop(loop_link, is_silent):
    if len(loop_link) == 2:
        return

    for sender, receiver in zip(loop_link[0:-1], loop_link[1:]):
        sender.send_message(receiver, is_silent)


def cal_marginal_from_tensor(prob_tensor, varnode_list):
    rtn_marginal = []
    for i, _ in enumerate(varnode_list):
        rtn_marginal.append(reduction_ndarray(prob_tensor, i))
    return rtn_marginal



import uuid
from abc import ABC, abstractmethod
import numpy as np
from cbp.utils.message import Message


class BaseNode(ABC):
    """All kinds node must inherit :class `~cbp.node.BaseNode`
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self, node_coef, potential) -> None:
        """Initialize default attr

        Every node need to have the following attr:

          * ``name`` str. id for the node

        :param node_coef: works for the norm-product algorithm
        :type node_coef: float
        :param potential: potential
        :type potential: ndarray or list
        """
        self.name = str(uuid.uuid4())
        self.node_coef = node_coef
        self._potential = None
        self.potential = potential
        self.epsilon = 1
        self.coef_ready = False
        self.is_traversed = False
        self.parent = None
        self.node_degree = 0
        self.connections = []
        self.message_inbox = {}
        self.latest_message = []
        self.connected_nodes = {}

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return self.__str__()

    @property
    def potential(self):
        return self._potential

    @potential.setter
    def potential(self, potential):
        self._potential = self._check_potential(potential)

    @abstractmethod
    def _check_potential(self, potential) -> np.ndarray:
        """check potential before set node potential

        :param potential: input potential
        :type potential: np.ndarray
        :return: [description]
        :rtype: np.ndarray
        """

    def format_name(self, name):
        self.name = name

    def reset_node_coef(self, coef):
        self.node_coef = coef

    def auto_coef(self, node_map, assign_policy=None):
        if assign_policy is None:
            self.node_coef = 1.0 / len(node_map)
        else:
            self.node_coef = assign_policy(self, node_map)

        self.register_nodes(node_map)

    # TODO:, should remove node_map parameter
    def cal_cnp_coef(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} is an abstract class")

    def check_before_run(self, node_map):
        for item in self.connections:
            assert item in node_map, f"{self.name} has a connection {item}, \
                                                    which is not in node_map"

    def make_init_message(self, recipient_node_name):
        if self.coef_ready:
            recipient_node = self.connected_nodes[recipient_node_name]
            message_dim = recipient_node.potential.shape
            return np.ones(message_dim)

        raise RuntimeError(
            f"Need to call cal_cnp_coef first for {self.name}")

    # keep all message looks urgly. convenient for debug and resource occupied
    # is not so huge
    def store_message(self, message):
        sender_name = message.sender.name
        self.message_inbox[sender_name] = message

        self.latest_message = list(self.message_inbox.values())

    def reset(self):
        self.message_inbox.clear()

    # TODO: FIXAPI NAME
    @abstractmethod
    def make_message(self, recipient_node) -> np.ndarray:
        """produce the val of message from current node to the recipient_node

        :param recipient_node: target node
        :type recipient_node: [type]
        :return: content of the message
        :rtype: np.ndarray
        """

    @abstractmethod
    def cal_bethe(self, margin) -> float:
        """calculate the bethe energy

        :return: bethe energy on this node
        :rtype: float
        """

    def send_message(self, recipient_node, is_silent=True):
        val = self.make_message(recipient_node)
        message = Message(self, val)
        recipient_node.store_message(message)
        if not is_silent:
            print(self.name + '->' + recipient_node.name)
            print(message.val)

    def sendin_message(self, is_silent=True):
        for connected_node in self.connected_nodes.values():
            connected_node.send_message(self, is_silent)

    def sendout_message(self, is_silent=True):
        for connected_node in self.connected_nodes.values():
            self.send_message(connected_node, is_silent)

    def register_connection(self, node_name):
        self.node_degree += 1
        self.connections.append(node_name)

    def register_nodes(self, node_map):
        for item in self.connections:
            if item in node_map:
                self.connected_nodes[item] = node_map[item]
            else:
                raise IOError(f"connection of {item} of {self.name} \
                                do not appear in the node_map")

    def get_connections(self):
        return self.connections

    def search_node_index(self, node_name):
        return self.connections.index(node_name)

    def search_msg_index(self, message_list, node_name):
        which_index = [i for i, message in enumerate(message_list)
                       if message.sender.name == node_name]
        if which_index:
            return which_index[0]

        raise RuntimeError(
            f"{node_name} do not appear in {self.name} message")



import unittest

import numpy as np

from .utils import two_node_tree


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.graph = two_node_tree()

    def test_sinkhorn_single_0(self):
        self.graph.get_node('VarNode_000').constrained_marginal \
            = np.array([0.5, 0.5])
        self.graph.get_node('VarNode_000').isconstrained = True
        self.graph.constrained_names.append('VarNode_000')
        self.graph.run_bp()
        node_equal = np.isclose(
            self.graph.get_node('VarNode_001').marginal(),
            np.array([0.375, 0.625])
        )
        self.assertTrue(all(node_equal))
        node_equal = np.isclose(
            self.graph.get_node('VarNode_000').marginal(),
            np.array([0.5, 0.5])
        )
        self.assertTrue(all(node_equal))

    def test_sinkhorn_single_1(self):
        self.graph.get_node('VarNode_001').constrained_marginal \
            = np.array([0.5, 0.5])
        self.graph.get_node('VarNode_001').isconstrained = True
        self.graph.constrained_names.append('VarNode_001')
        self.graph.run_bp()
        node_equal = np.isclose(
            self.graph.get_node('VarNode_001').marginal(),
            np.array([0.5, 0.5])
        )
        self.assertTrue(all(node_equal))
        node_equal = np.isclose(
            self.graph.get_node('VarNode_000').marginal(),
            np.array([23.0 / 76, 53.0 / 76])
        )
        self.assertTrue(all(node_equal))

    def test_sinkhorn_2(self):
        self.graph.get_node('VarNode_000').constrained_marginal \
            = np.array([0.5, 0.5])
        self.graph.get_node('VarNode_000').isconstrained = True
        self.graph.get_node('VarNode_001').constrained_marginal \
            = np.array([0.5, 0.5])
        self.graph.get_node('VarNode_001').isconstrained = True
        self.graph.constrained_names.append('VarNode_000')
        self.graph.constrained_names.append('VarNode_001')
        self.graph.run_bp()
        node_equal = np.isclose(
            self.graph.get_node('VarNode_000').marginal(),
            np.array([0.5, 0.5])
        )
        self.assertTrue(all(node_equal))
        node_equal = np.isclose(
            self.graph.get_node('VarNode_001').marginal(),
            np.array([0.5, 0.5])
        )
        self.assertTrue(all(node_equal))



import numpy as np
import hmmlearn.hmm as hmm

# pylint: skip-file

class HMMEngine(hmm.MultinomialHMM):
    def __init__(self, n_components=1,
                 startprob_prior=1.0, transmat_prior=1.0,
                 algorithm="viterbi", random_state=None,
                 n_iter=10, tol=1e-2, verbose=False,
                 params="ste", init_params="ste"):
        hmm.MultinomialHMM.__init__(self, n_components,
                                    startprob_prior=startprob_prior,
                                    transmat_prior=transmat_prior,
                                    algorithm=algorithm,
                                    random_state=random_state,
                                    n_iter=n_iter, tol=tol, verbose=verbose,
                                    params=params, init_params=init_params)

    def set_simprob(self, hmm_simulator):
        self.startprob_ = hmm_simulator.get_init_potential()
        self.transmat_ = hmm_simulator.get_transition_potential()
        self.emissionprob_ = hmm_simulator.get_emission_potential()

    def set_dictprob(self, dict_prob):
        self.startprob_ = dict_prob.pi
        self.transmat_ = dict_prob.transition
        self.emissionprob_ = dict_prob.emission

    def score(self, traj, time_length):
        num_seq = traj.size // time_length
        length = [time_length] * num_seq
        return super().score(traj, length) / num_seq

    def fit(self, traj, time_length):
        num_seq = traj.size // time_length
        length = [time_length] * num_seq
        super().fit(traj, length)

    def register_init(self, dict_prob):
        self.init_param = dict_prob

    def _init(self, X, lengths=None):
        self.set_dictprob(self.init_param)



import unittest

from cbp.graph.coef_policy import avg_policy, bp_policy

from .utils import six_node_graph


class TestSetCNPCoef(unittest.TestCase):
    def setUp(self):
        self.graph = six_node_graph()

    def test_avg_policy(self):
        self.graph.coef_policy = avg_policy
        self.graph.init_node_recorder()
        self.graph.bake()
        for node in self.graph.nodes:
            node.cal_cnp_coef()

        for node in self.graph.nodes:
            self.assertAlmostEqual(node.node_coef, 1.0 / 11)
        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_000").get_hat_c_ialpha("VarNode_000"), 2.0 / 11)
        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_000").get_hat_c_ialpha("VarNode_001"), 10.0 / 11)
        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_002").get_hat_c_ialpha("VarNode_001"), 6.0 / 11)
        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_004").get_hat_c_ialpha("VarNode_005"), 2.0 / 11)

    def test_bp_policy(self):
        self.graph.coef_policy = bp_policy
        self.graph.init_node_recorder()
        self.graph.bake()
        for node in self.graph.nodes:
            node.cal_cnp_coef()

        self.assertAlmostEqual(
            self.graph.get_node("VarNode_001").node_coef, -2)
        self.assertAlmostEqual(
            self.graph.get_node("VarNode_003").node_coef, -2)

        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_000").get_hat_c_ialpha("VarNode_000"), 1)
        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_000").get_hat_c_ialpha("VarNode_001"), 1)
        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_002").get_hat_c_ialpha("VarNode_001"), 1)
        self.assertAlmostEqual(self.graph.get_node(
            "FactorNode_004").get_hat_c_ialpha("VarNode_005"), 1)



import re
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the version string
with open(path.join(here, 'cbp', '__init__.py')) as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name='pycbp',
    version=version,
    description='A Library for Constrained Belief Propagation',
    long_description=open('README.md', encoding='utf8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/zqsh/cbp',
    author='Qinsheng',
    author_email='qsh.zh27@gmail.com',
    license='MIT',
    python_requires='>=3.7',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='belief propagation bayesian inference',
    packages=find_packages(exclude=['test', 'test.*',
                                    'examples', 'examples.*',
                                    'docs', 'docs.*']),
    install_requires=[
        'numpy',
        'tqdm',
        'scipy',
        'matplotlib',
        'seaborn',
        'pygraphviz',
        'numba',
        'hmmlearn'
    ],
    extras_require={
        'dev': [
            'Sphinx',
            'sphinx_rtd_theme',
            'nbsphinx',
            'sphinxcontrib-bibtex',
            'pylint',
            'pytest',
            'pytest-cov',
        ]
    },
)



from .base_builder import BaseBuilder


class LineBuilder(BaseBuilder):
    def __init__(self, num_node, d, policy, rand_seed=1):
        self.num_node = num_node
        super().__init__(d, policy, rand_seed)

    def init_graph(self):
        self.add_constrained_node()
        for _ in range(self.num_node - 2):
            self.add_branch(is_constrained=False)
        self.add_branch(is_constrained=True)
